/*--------------------------------------------------------------------------+
$Id: codetemplates.xml 1 2011-01-01 00:00:01Z hoelzl $
|                                                                          |
| Copyright 2015 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editpart.allocation;

import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_SHAPE_INSETS;
import static org.fortiss.tooling.base.ui.editpart.ExtendedLayerRootEditPart.DECORATION_LAYER;
import static org.fortiss.tooling.base.ui.editpart.ExtendedLayerRootEditPart.LABEL_LAYER;
import static org.fortiss.tooling.base.ui.utils.DecorationIconUtils.getMarkerServiceDecorationIcon;
import static org.fortiss.tooling.kernel.utils.KernelModelElementUtils.getParentElement;

import java.util.ArrayList;
import java.util.List;

import org.conqat.ide.commons.ui.swt.FontUtils;
import org.eclipse.draw2d.IFigure;
import org.eclipse.draw2d.Label;
import org.eclipse.draw2d.PositionConstants;
import org.eclipse.draw2d.geometry.Insets;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.gef.EditPart;
import org.eclipse.gef.NodeEditPart;
import org.eclipse.gef.Request;
import org.eclipse.gef.RequestConstants;
import org.eclipse.gef.RootEditPart;
import org.eclipse.jface.resource.JFaceResources;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.fortiss.tooling.base.layout.DefaultLayoutConstants;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.ui.editpart.GraphicalEditPartBase;
import org.fortiss.tooling.base.ui.editpart.PositionedEditPartBase;
import org.fortiss.tooling.base.ui.editpart.figure.PrettyRoundedRectangle;
import org.fortiss.tooling.base.ui.editpart.figure.TransparentLabel;
import org.fortiss.tooling.base.ui.layout.IDiagramLayoutConfiguration;
import org.fortiss.tooling.kernel.extension.data.IConstraintViolation.ESeverity;
import org.fortiss.tooling.kernel.model.INamedCommentedElement;
import org.fortiss.tooling.kernel.ui.service.IMarkerService;
import org.fortiss.tooling.kernel.ui.service.IModelElementHandlerService;

/**
 * Implementation similar to {@link PositionedEditPartBase} except that <b>T</b> does not have to be
 * a {@link ILayoutedModelElement}. <b>S</b> is the type of model element to be displayed in this
 * sub diagram.
 * 
 * @author eder
 * @author $Author: hoelzl $
 * @version $Rev: 18709 $
 * @param <T>
 * @ConQAT.Rating RED Hash:
 */
public abstract class AllocationSubDiagramEditPartBase<T extends INamedCommentedElement> extends
		GraphicalEditPartBase<T> implements NodeEditPart {

	/** The base figure of this positioned edit part. */
	protected IFigure baseFigure;

	/** The label figure. */
	protected Label labelFigure;

	/** The decoration figure. */
	protected Label decorationFigure;

	/** The layout configuration. */
	private IDiagramLayoutConfiguration layoutConfiguration;

	/** List of all children visible in this edit part. */
	private List<?> modelChildren = new ArrayList<>();
	/** The default font for titles. */
	public static Font DEFAULT_TITLE_FONT = FontUtils.derive(JFaceResources.getDefaultFont(), 0,
			SWT.BOLD);

	/**
	 * @param modelElement
	 */
	protected AllocationSubDiagramEditPartBase(T modelElement,
			IDiagramLayoutConfiguration layoutConfiguration) {
		super(modelElement);
		this.layoutConfiguration = layoutConfiguration;
		modelChildren = getInitialModelSubElements();
	}

	/** {@inheritDoc} */
	@Override
	protected IFigure createFigure() {
		PrettyRoundedRectangle prettyRoundedRectangle =
				new PrettyRoundedRectangle(0, new Insets(
						DefaultLayoutConstants.DEFAULT_CONNECTOR_SIZE / 2 - 2)) {
					/** {@inheritDoc} */
					@Override
					protected boolean useLocalCoordinates() {
						return true;
					}
				};
		baseFigure = prettyRoundedRectangle;
		return prettyRoundedRectangle;
	}

	/** Creates the label figure. */
	public Label createLabelFigure() {
		Label nameLabel = new TransparentLabel();
		nameLabel.setForegroundColor(getForegroundColor());
		nameLabel.setFont(DEFAULT_TITLE_FONT);
		nameLabel.setTextAlignment(PositionConstants.LEFT | PositionConstants.TOP);
		nameLabel.setTextPlacement(PositionConstants.EAST);
		nameLabel.setIcon(IModelElementHandlerService.INSTANCE.getModelElementHandler(getModel())
				.getIcon(getModel()));
		return nameLabel;
	}

	/**
	 * Creates the decoration figure. The default implementation does not create
	 * a figure.
	 */
	protected Label createDecorationFigure() {
		return new TransparentLabel();
	}

	/** {@inheritDoc} */
	@Override
	public void addNotify() {
		labelFigure = createLabelFigure();
		if(labelFigure != null) {
			registerLabelFigure(labelFigure);
		}
		decorationFigure = createDecorationFigure();
		if(decorationFigure != null) {
			registerDecorationFigure(decorationFigure);
		}
		// called last, since it also refreshes label and decoration
		super.addNotify();
	}

	/** {@inheritDoc} */
	@Override
	public void removeNotify() {
		if(labelFigure != null && labelFigure.getParent() != null) {
			labelFigure.getParent().remove(labelFigure);
		}
		if(decorationFigure != null && decorationFigure.getParent() != null) {
			decorationFigure.getParent().remove(decorationFigure);
		}
		super.removeNotify();
	}

	/**
	 * Register the new label figure into the label figure structure.
	 */
	protected void registerLabelFigure(IFigure figure) {
		if(getParent() instanceof AllocationSubDiagramEditPartBase)
			((AllocationSubDiagramEditPartBase<?>)getParent()).registerLabelFigure(figure);
		else
			getLayer(LABEL_LAYER).add(figure);
	}

	/**
	 * Register the new decoration figure into the decoration figure structure.
	 */
	protected void registerDecorationFigure(IFigure figure) {
		if(getParent() instanceof AllocationSubDiagramEditPartBase)
			((AllocationSubDiagramEditPartBase<?>)getParent()).registerDecorationFigure(figure);
		else
			getLayer(DECORATION_LAYER).add(figure);
	}

	/** {@inheritDoc} */
	@Override
	protected final List<?> getModelChildren() {
		return modelChildren;
	}

	/** Returns a list of all sub elements contained by T */
	protected abstract List<?> getInitialModelSubElements();

	/**
	 * Returns the position and size of the background figure of this sub diagram.
	 */
	protected Rectangle determineBaseFigureBounds() {
		RootEditPart root = getRoot();
		Rectangle sourceBounds =
				IAllocationPositionService.INSTANCE.getSourceAllocationBounds(root);
		Rectangle targetBounds =
				IAllocationPositionService.INSTANCE.getTargetAllocationBounds(root);
		Point minTargetPos = IAllocationPositionService.INSTANCE.getMinTargetPos(root);
		if(isTargetDiagram()) {
			return new Rectangle(targetBounds.x, sourceBounds.y + sourceBounds.height,
					targetBounds.width + minTargetPos.x, targetBounds.height + 20);
		}
		return new Rectangle(sourceBounds.x, sourceBounds.y, sourceBounds.width,
				sourceBounds.height);
	}

	/**
	 * Returns the position and size of the decoration figure. The computation
	 * may use the bounds of the base figure or the label figure, since both
	 * bounds were determined first.
	 */
	protected Rectangle determineDecorationFigureBounds() {
		Rectangle r = determineBaseFigureBounds();
		r.x += 4;
		r.y += 4;
		r.width = useSmallDecorationImage() ? 8 : 16;
		r.height = useSmallDecorationImage() ? 8 : 16;
		return r;
	}

	/**
	 * Returns whether to use small or big decoration image. Default
	 * implementation uses big image.
	 */
	protected boolean useSmallDecorationImage() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	protected void createEditPolicies() {
		// nothing to do
	}

	/** Returns the foreground color used. */
	protected abstract Color getForegroundColor();

	/** Returns the background color used. */
	protected abstract Color getBackgroundColor();

	/**
	 * Returns if this edit part is the target edit part of the allocation.
	 */
	abstract public boolean isTargetDiagram();

	/** {@inheritDoc} */
	@Override
	protected void refreshVisuals() {
		if(baseFigure != null && baseFigure.getParent() != null) {
			Rectangle baseBounds = determineBaseFigureBounds();
			baseFigure.setBounds(baseBounds);
			baseFigure.getParent().setConstraint(baseFigure, baseBounds);
			baseFigure.setForegroundColor(getForegroundColor());
			baseFigure.setBackgroundColor(getBackgroundColor());
			baseFigure.setToolTip(new Label(getBaseFigureTooltip()));
		}

		if(labelFigure != null && labelFigure.getParent() != null) {
			labelFigure.setText(determineLabelFigureText());
			Rectangle labelBounds = determineLabelFigureBounds();
			labelFigure.setBounds(labelBounds);
			labelFigure.getParent().setConstraint(labelFigure, labelBounds);
		}

		if(decorationFigure != null && decorationFigure.getParent() != null) {
			refreshDecorationFigure();
		}

	}

	/**
	 * Returns the position and size of the label figure. The text of the label
	 * has been updated using {@link #determineLabelFigureText()} just before
	 * this method is called. The computation may also use the bounds of the
	 * base figure, since its bounds were determined first. The default
	 * implementation returns an invisible rectangle.
	 */
	protected Rectangle determineLabelFigureBounds() {
		Rectangle textBounds = labelFigure.getTextBounds();
		textBounds.y = DEFAULT_SHAPE_INSETS / 2;
		textBounds.x = DEFAULT_SHAPE_INSETS;
		textBounds.width = textBounds.width + 4 * DEFAULT_SHAPE_INSETS;
		textBounds.height = textBounds.height * 2;
		if(isTargetDiagram()) {
			textBounds.y += baseFigure.getBounds().y;
		}
		return textBounds;
	}

	/**
	 * Determines the text of the label figure during a refresh. This method is
	 * called right before {@link #determineLabelFigureBounds()}. The default
	 * implementation returns an empty string.
	 */
	protected String determineLabelFigureText() {
		return getModel().getName();
	}

	/** Returns the current tooltip description. */
	protected String getBaseFigureTooltip() {
		return "";
	}

	/** Refreshes the decoration figure using {@link IMarkerService}. */
	protected void refreshDecorationFigure() {
		Rectangle decorationBounds = determineDecorationFigureBounds();
		decorationFigure.setBounds(decorationBounds);
		if(decorationFigure.getParent() != null && decorationBounds != null) {
			decorationFigure.getParent().setConstraint(decorationFigure, decorationBounds);

			ESeverity severity = IMarkerService.INSTANCE.getHighestViolationSeverity(getModel());
			decorationFigure.setIcon(getMarkerServiceDecorationIcon(severity,
					useSmallDecorationImage()));
		}
	}

	/** {@inheritDoc} */
	@Override
	public final IDiagramLayoutConfiguration getLayoutConfiguration() {
		return layoutConfiguration;
	}

	/** {@inheritDoc} */
	@Override
	public void performRequest(Request req) {
		if(req.getType() == RequestConstants.REQ_OPEN) {
			List<Object> children = new ArrayList<>();
			if(req.getExtendedData().isEmpty()) {
				// sub diagram background selection
				// works only with hierarchic elements
				if(!modelChildren.isEmpty()) {
					Object object = modelChildren.get(0);
					IHierarchicElement parentElement =
							getParentElement((EObject)object, IHierarchicElement.class, false);
					if(parentElement == null) {
						return;
					}
					IHierarchicElement parentParentElement =
							getParentElement(parentElement, IHierarchicElement.class, false);
					if(parentParentElement != null) {
						children.addAll(parentParentElement.getContainedElements());
						children.addAll(parentParentElement.getConnectors());
					}
				}
			}
			Object object = req.getExtendedData().get(this);
			if(object instanceof List<?> && !((List<?>)object).isEmpty()) {
				// sub diagram child selection
				children.addAll((List<?>)object);
			}
			// remove all child edit parts and set the new model children. after the refresh the
			// new child edit parts are available.
			clearChildrenEditParts();
			modelChildren = children;
			refresh();
		}
	}

	/** Remove all child edit parts */
	private void clearChildrenEditParts() {
		List<?> list = getChildren();
		for(int i = 0; i < list.size(); i++) {
			((EditPart)list.get(i)).removeNotify();
		}
	}

}
